
#include "GLObject.H"

//#define DONOT_USE_GL_1_1_VERTEX_ARRAY

// OpenGL Extensions pointer object
#ifdef WIN32
//GLExtensionsProcPointer GLExt ;

GLExtensionsProcPointer::GLExtensionsProcPointer()
{
	// 拡張子リスト
	const char *ext = (const char *)::glGetString(GL_EXTENSIONS) ;
	extensions = ext ;

	// SGIS_multitexture
	if (strstr(ext, "GL_SGIS_multitexture"))
	{
		SGIS_multitexture = TRUE ;

		// get SGIS_multitexture procedure pointers
		glSelectTextureCoordSetSGIS = wglGetProcAddress("glSelectTextureCoordSetSGIS") ;
		glSelectTextureSGIS         = wglGetProcAddress("glSelectTextureSGIS") ;
		glMTexCoordPointerSGIS      = wglGetProcAddress("glMTexCoordPointerSGIS") ;
		glMTexCoord1fSGIS           = wglGetProcAddress("glMTexCoord1fSGIS") ;
		glMTexCoord1fvSGIS          = wglGetProcAddress("glMTexCoord1fvSGIS") ;
		glMTexCoord2fSGIS           = wglGetProcAddress("glMTexCoord2fSGIS") ;
		glMTexCoord2fvSGIS          = wglGetProcAddress("glMTexCoord2fvSGIS") ;
		glMTexCoord3fSGIS           = wglGetProcAddress("glMTexCoord3fSGIS") ;
		glMTexCoord3fvSGIS          = wglGetProcAddress("glMTexCoord3fvSGIS") ;
	}
}
#endif

// イメージフォーマットを RGB から RGBA に（RGBA フォーマットしか使えない環境ため）
// アルファ値のデフォルトは 1.0f
int RGBtoRGBA(TRUEIMAGE *image, GLfloat alpha)
{
	unsigned int x, y ;
	unsigned char *newImage ;

	if (!(newImage = (unsigned char *)malloc(image->width * image->height * 4))) return FAILURE ;

	if      (alpha < 0.0f) alpha = 0.0f ;
	else if (alpha > 0.0f) alpha = 1.0f ;

	for (y = 0 ; y < image->height ; y ++)
	{
		for (x = 0 ; x < image->width ; x ++)
		{
			unsigned char *ptr = image->dataptr + y * image->width * 3 + x * 3,	// 元イメージポインタ
						  *npt = newImage + y * image->width * 4 + x * 4 ;		// RGBAイメージポインタ

			npt[0] = ptr[0] ; npt[1] = ptr[1] ; npt[2] = ptr[2] ;

			npt[3] = (char)(alpha * 255.0f) ;	// アルファ値は 1.0f
		}
	}

	free(image->dataptr) ;
	image->dataptr = newImage ;

	return SUCCESS ;
}

// イメージガンマ補正
void GammaCollect(TRUEIMAGE *image, float gamma, unsigned int ix, unsigned int iy, unsigned int width, unsigned int height)
{
	// 変化なしなら即リターン
	if (gamma == 1.0f) return ;

	unsigned int x, y ;

	if (gamma == 0.0f) return ;
	else gamma = 1.0f / gamma ;

	width  += ix ;
	height += iy ;

	if (width  == ix || width  > image->width ) width  = image->width ;
	if (height == iy || height > image->height) height = image->height ;

	for (y = iy ; y < height ; y ++)
	{
		for (x = ix ; x < width ; x ++)
		{
			unsigned char *ptr = image->dataptr + y * image->width * 3 + x * 3 ;	// R 要素のポインタ

			ptr[0] = (unsigned char)(Pow(ptr[0] / 255.0f, gamma) * 255.0f) ;	// R 補正
			ptr[1] = (unsigned char)(Pow(ptr[1] / 255.0f, gamma) * 255.0f) ;	// G 補正
			ptr[2] = (unsigned char)(Pow(ptr[2] / 255.0f, gamma) * 255.0f) ;	// B 補正
		}
	}
}

// イメージ明るさ変化
void ChangeBright(TRUEIMAGE *image, float bright, unsigned int ix, unsigned int iy, unsigned int width, unsigned int height)
{
	// 変化なしなら即リターン
	if (bright == 1.0f) return ;

	unsigned int x, y ;

	width  += ix ;
	height += iy ;

	if (width  == ix || width  > image->width ) width  = image->width ;
	if (height == iy || height > image->height) height = image->height ;

	for (y = iy ; y < height ; y ++)
	{
		for (x = ix ; x < width ; x ++)
		{
			unsigned char *ptr = image->dataptr + y * image->width * 3 + x * 3 ;	// R 要素のポインタ

			ptr[0] = (unsigned char)(Clump(ptr[0] * bright, 0, 255.0f)) ;	// R 補正
			ptr[1] = (unsigned char)(Clump(ptr[1] * bright, 0, 255.0f)) ;	// G 補正
			ptr[2] = (unsigned char)(Clump(ptr[2] * bright, 0, 255.0f)) ;	// B 補正
		}
	}
}


// イメージの上下左右反転
void ReverseCoordinates(TRUEIMAGE *image, int mirrorFlag, int flipFlag)
{
	// 変化なしなら即リターン
	if (!mirrorFlag && !flipFlag) return ;

	// 作業用一時バッファ作成
	TRUEIMAGE *tmp ;
	tmp = (TRUEIMAGE *)malloc(sizeof(TRUEIMAGE)) ;
	if (!tmp) return ;
	tmp->dataptr = (unsigned char *)malloc(sizeof(unsigned char) * image->width * image->height * 3) ;
	if (!tmp->dataptr)
	{
		free(tmp) ;
		return ;
	}
	// データ部コピー
	memcpy(tmp->dataptr, image->dataptr, sizeof(unsigned char) * image->width * image->height * 3) ;

	// tmp を元に 座標反転しながらデータをコピー
	for (int y = 0 ; y < image->height ; y ++)
	{
		int sy ;
		if (flipFlag) sy = image->height - y - 1 ;
		else		  sy = y ;

		for (int x = 0 ; x < image->width ; x ++)
		{
			int sx ;
			if (mirrorFlag) sx = image->width - x - 1 ;
			else			sx = x ;

			unsigned char *srcptr =   tmp->dataptr + sy * image->width * 3 + sx * 3 ;	// R 要素のポインタ
			unsigned char *dstptr = image->dataptr +  y * image->width * 3 +  x * 3 ;	// R 要素のポインタ

			dstptr[0] = srcptr[0] ;
			dstptr[1] = srcptr[1] ;
			dstptr[2] = srcptr[2] ;
		}
	}

	freeimage(tmp) ;
}

// 視野ピラミッド作成（ジッタ情報含）
void Perspective(GLfloat scr_w, GLfloat scr_h, GLfloat n_clip, GLfloat f_clip,
				 int win_w, int win_h, GLfloat px, GLfloat py)
{
	if (win_w < 0 || win_h < 0)
	{
		GLint vp[4] ;
		glGetIntegerv(GL_VIEWPORT, vp) ;
		win_w = vp[2] ;
		win_h = vp[3] ;
	}

	px *= scr_w / win_w ;
	py *= scr_h / win_h ;

	scr_w *= 0.5f ;
	scr_h *= 0.5f ;

	glMatrixMode(GL_PROJECTION) ;
	glLoadIdentity() ;
	glFrustum(-scr_w + px, scr_w + px, -scr_h + py, scr_h + py, n_clip, f_clip) ;
}

// 色、素材セット
void SetMaterial(GLenum surface, GLfloat r, GLfloat g, GLfloat b, GLfloat a, GLfloat metal, GLfloat specular, GLfloat shininess, GLint mirror, GLfloat m_ratio, GLfloat amb, GLenum color_switch)
{
	GLfloat param[3][4], diffuse = 1.0f - metal, d_r, d_g, d_b ;
	if (mirror) diffuse = 1.0f - metal * (1.0f - m_ratio) ;

	param[0][0] = param[2][0] = d_r = r * diffuse ;
	param[0][1] = param[2][1] = d_g = g * diffuse ;
	param[0][2] = param[2][2] = d_b = b * diffuse ;
	param[0][3] = a ;

/*
#ifdef WIN32
	param[1][1] = (r + (1.0f - d_r) * (1.0f - metal)) * specular ;	// VC++ 最適化かけるとスペキュラーの r,g が逆？？？
	param[1][0] = (g + (1.0f - d_g) * (1.0f - metal)) * specular ;
	param[1][2] = (b + (1.0f - d_b) * (1.0f - metal)) * specular ;
	param[1][3] = 1.0f ;
#else
	param[1][0] = (r + (1.0f - d_r) * (1.0f - metal)) * specular ;
	param[1][1] = (g + (1.0f - d_g) * (1.0f - metal)) * specular ;
	param[1][2] = (b + (1.0f - d_b) * (1.0f - metal)) * specular ;
	param[1][3] = 1.0f ;
#endif
*/
	param[1][0] = (r + (1.0f - d_r) * (1.0f - metal)) * specular ;
	param[1][1] = (g + (1.0f - d_g) * (1.0f - metal)) * specular ;
	param[1][2] = (b + (1.0f - d_b) * (1.0f - metal)) * specular ;
	param[1][3] = 1.0f ;

	if (amb != NONE)
	{
		param[2][0] = r * amb ;
		param[2][1] = g * amb ;
		param[2][2] = b * amb ;
	}

	if (color_switch != NONE)
	{
		if (color_switch != GL_SPECULAR && (amb == NONE || color_switch != GL_AMBIENT)) glColor4fv(param[0]) ;
		else if (color_switch == GL_AMBIENT) glColor3fv(param[2]) ;
		else glColor3fv(param[1]) ;
	}
	else
	{
		if (amb == NONE) glMaterialfv(surface, GL_AMBIENT_AND_DIFFUSE, param[0]) ;
		else
		{
			glMaterialfv(surface, GL_DIFFUSE, param[0]) ;
			glMaterialfv(surface, GL_AMBIENT, param[2]) ;
		}
		glMaterialfv(surface, GL_SPECULAR,  param[1]) ;
		glMaterialf (surface, GL_SHININESS, shininess) ;
	}
}

void SetLight(GLenum light, GLfloat pos[4], GLfloat color[4], GLfloat amb, GLint unitFlag, GLint specularFlag)
{
	GLfloat	lamb[4] = { color[0] * amb, color[1] * amb, color[2] * amb, 1.0f } ;
	static GLfloat	u_pos[4] ;
	static GLfloat	spec[4] = { 0.0f, 0.0f, 0.0f, 1.0f } ;

	Vector3d u(pos) ;
	if (unitFlag) u.Unit() ;

	u_pos[0] = u.X() ;
	u_pos[1] = u.Y() ;
	u_pos[2] = u.Z() ;
	u_pos[3] = pos[3] ;

	glLightfv(light, GL_POSITION, u_pos) ;
	glLightfv(light, GL_AMBIENT,  lamb  ) ;
	glLightfv(light, GL_DIFFUSE,  color) ;

	if (specularFlag) glLightfv(light, GL_SPECULAR, color) ;
	else              glLightfv(light, GL_SPECULAR, spec) ;
}

// 文字列描画
void DrawString(String& str, GLfloat x, GLfloat y, GLfloat z, void *font)
{
	glRasterPos3f(x, y, z) ;

	for (int i = 1 ; i <= str.Length() ; i ++)
		glutBitmapCharacter(font, str.Character(i)) ;
}

// 文字列描画
void DrawString(String& str, void *font)
{
	for (int i = 1 ; i <= str.Length() ; i ++)
		glutBitmapCharacter(font, str.Character(i)) ;
}

// サンプル・ジッタ生成
Vector2d *CreateSampleJitter(int sample)
{
	Vector2d *jitter ;

	switch (sample)
	{
	case 0:
	case 1:
	case 2:
		jitter = new Vector2d[2] ;
		jitter[0](0.25f, 0.75f) ;
		jitter[1](0.75f, 0.25f) ;
		break ;
	case 3:
		jitter = new Vector2d[3] ;
		jitter[0](0.5033922635f, 0.8317967229f) ;
		jitter[1](0.7806016275f, 0.2504380877f) ;
		jitter[2](0.2261828938f, 0.4131553612f) ;
		break ;
	case 4:
		jitter = new Vector2d[4] ;
		jitter[0](0.375f, 0.25f) ;
		jitter[1](0.125f, 0.75f) ;
		jitter[2](0.875f, 0.25f) ;
		jitter[3](0.625f, 0.75f) ;
		break ;
	case 5:
		jitter = new Vector2d[5] ;
		jitter[0](0.5f, 0.5f) ;
		jitter[1](0.3f, 0.1f) ;
		jitter[2](0.7f, 0.9f) ;
		jitter[3](0.9f, 0.3f) ;
		jitter[4](0.1f, 0.7f) ;
		break ;
	case 6:
		jitter = new Vector2d[6] ;
		jitter[0](0.4646464646f, 0.4646464646f) ;
		jitter[1](0.1313131313f, 0.7979797979f) ;
		jitter[2](0.5353535353f, 0.8686868686f) ;
		jitter[3](0.8686868686f, 0.5353535353f) ;
		jitter[4](0.7979797979f, 0.1313131313f) ;
		jitter[5](0.2020202020f, 0.2020202020f) ;
		break ;
	case 7:
	case 8:
		jitter = new Vector2d[8] ;
		jitter[0](0.5625f, 0.4375f) ;
		jitter[1](0.0625f, 0.9375f) ;
		jitter[2](0.3125f, 0.6875f) ;
		jitter[3](0.6875f, 0.8125f) ;
		jitter[4](0.8125f, 0.1875f) ;
		jitter[5](0.9375f, 0.5625f) ;
		jitter[6](0.4375f, 0.0625f) ;
		jitter[7](0.1875f, 0.3125f) ;
		break ;
	case 9:
		jitter = new Vector2d[9] ;
		jitter[0](0.5f         , 0.5f         ) ;
		jitter[1](0.1666666666f, 0.9444444444f) ;
		jitter[2](0.5f         , 0.1666666666f) ;
		jitter[3](0.5f         , 0.8333333333f) ;
		jitter[4](0.1666666666f, 0.2777777777f) ;
		jitter[5](0.8333333333f, 0.3888888888f) ;
		jitter[6](0.1666666666f, 0.6111111111f) ;
		jitter[7](0.8333333333f, 0.7222222222f) ;
		jitter[8](0.8333333333f, 0.0555555555f) ;
		break ;
	case 10:
	case 11:
	case 12:
		jitter = new Vector2d[12] ;
		jitter[ 0](0.4166666666f, 0.625f) ;
		jitter[ 1](0.9166666666f, 0.875f) ;
		jitter[ 2](0.25f        , 0.375f) ;
		jitter[ 3](0.4166666666f, 0.125f) ;
		jitter[ 4](0.75f        , 0.125f) ;
		jitter[ 5](0.0833333333f, 0.125f) ;
		jitter[ 6](0.75f        , 0.625f) ;
		jitter[ 7](0.25f        , 0.875f) ;
		jitter[ 8](0.5833333333f, 0.375f) ;
		jitter[ 9](0.9166666666f, 0.375f) ;
		jitter[10](0.0833333333f, 0.625f) ;
		jitter[11](0.5833333333f, 0.875f) ;
		break ;
	case 13:
	case 14:
	case 15:
	case 16:
		jitter = new Vector2d[16] ;
		jitter[ 0](0.375f, 0.4375f) ;
		jitter[ 1](0.625f, 0.0625f) ;
		jitter[ 2](0.875f, 0.1875f) ;
		jitter[ 3](0.125f, 0.0625f) ;
		jitter[ 4](0.375f, 0.6875f) ;
		jitter[ 5](0.875f, 0.4375f) ;
		jitter[ 6](0.625f, 0.5625f) ;
		jitter[ 7](0.375f, 0.9375f) ;
		jitter[ 8](0.625f, 0.3125f) ;
		jitter[ 9](0.125f, 0.5625f) ;
		jitter[10](0.125f, 0.8125f) ;
		jitter[11](0.375f, 0.1875f) ;
		jitter[12](0.875f, 0.9375f) ;
		jitter[13](0.875f, 0.6875f) ;
		jitter[14](0.125f, 0.3125f) ;
		jitter[15](0.625f, 0.8125f) ;
		break ;
	
	default:
		return NULL ;
	}
	return jitter ;
}

void DeleteSampleJitter(Vector2d *jitter)
{
	if (jitter) delete [] jitter ;
}

Vector3d& Vector3d::Unit()
{
	GLfloat len = sqrt(X()*X() + Y()*Y() + Z()*Z()) ;
	if (len != 0.0f) *this /= len ;
	return *this ;
}

/*
const Vector3d Unit(Vector3d& vec)
{
	return (Vector3d(vec).Unit()) ;
}
*/
Vector3d Unit(const Vector3d& vec)
{
	return (Vector3d(vec).Unit()) ;
}

/*
const Vector3d UnitNormal(Vector3d& v1, Vector3d& v2)
{
	return Unit(v1 | v2) ;
}
const Vector3d UnitNormal(Vector3d& v1, const Vector3d& v2)
{
	return Unit(v1 | v2) ;
}
const Vector3d UnitNormal(const Vector3d& v1, Vector3d& v2)
{
	return Unit(v1 | v2) ;
}
*/
Vector3d UnitNormal(const Vector3d& v1, const Vector3d& v2)
{
	return Unit(v1 | v2) ;
}

/*
const Vector3d operator |(Vector3d& v1, Vector3d& v2)
{
	return Vector3d(v1.Y()*v2.Z() - v1.Z()*v2.Y(),
			v1.Z()*v2.X() - v1.X()*v2.Z(),
			v1.X()*v2.Y() - v1.Y()*v2.X()) ;
}
const Vector3d operator |(Vector3d& v1, const Vector3d& v2)
{
	return Vector3d(v1.Y()*v2.Z() - v1.Z()*v2.Y(),
			v1.Z()*v2.X() - v1.X()*v2.Z(),
			v1.X()*v2.Y() - v1.Y()*v2.X()) ;
}
const Vector3d operator |(const Vector3d& v1, Vector3d& v2)
{
	return Vector3d(v1.Y()*v2.Z() - v1.Z()*v2.Y(),
			v1.Z()*v2.X() - v1.X()*v2.Z(),
			v1.X()*v2.Y() - v1.Y()*v2.X()) ;
}
*/
Vector3d operator |(const Vector3d& v1, const Vector3d& v2)
{
	return Vector3d(v1.Y()*v2.Z() - v1.Z()*v2.Y(),
			v1.Z()*v2.X() - v1.X()*v2.Z(),
			v1.X()*v2.Y() - v1.Y()*v2.X()) ;
}

// 行列で座標変換
Vector3d& Vector3d::Transform(const GLfloat m[16])
{
	int      i ;
	for (i = 0 ; i < 3 ; i ++)
		v[i] = m[i]*v[0] + m[4+i]*v[1] + m[8+i]*v[2] + m[12+i] ;
	return *this ;
}

// 行列で座標変換（回転／拡大縮小部分のみの演算 法線ベクトル変換用）
Vector3d& Vector3d::Transform3x3(const GLfloat m[16])
{
	int      i ;
	for (i = 0 ; i < 3 ; i ++)
		v[i] = m[i]*v[0] + m[4+i]*v[1] + m[8+i]*v[2] ;
	return *this ;
}

ostream& operator <<(ostream& os, const Vector3d& vec)
{
	os << vec.X() << ", " << vec.Y() << ", " << vec.Z() ;
	return os ;
}


Object& Object::MakePolygon(int polNo, int v1, int v2, int v3, const GLfloat data[][3], const GLfloat m[16])
{
	polygon[polNo].vertex[0].X() = data[v1][0] ;
	polygon[polNo].vertex[0].Y() = data[v1][1] ;
	polygon[polNo].vertex[0].Z() = data[v1][2] ;
	if (m) polygon[polNo].vertex[0].Transform(m) ;

	polygon[polNo].vertex[1].X() = data[v2][0] ;
	polygon[polNo].vertex[1].Y() = data[v2][1] ;
	polygon[polNo].vertex[1].Z() = data[v2][2] ;
	if (m) polygon[polNo].vertex[1].Transform(m) ;

	polygon[polNo].vertex[2].X() = data[v3][0] ;
	polygon[polNo].vertex[2].Y() = data[v3][1] ;
	polygon[polNo].vertex[2].Z() = data[v3][2] ;
	if (m) polygon[polNo].vertex[2].Transform(m) ;

	polygon[polNo].pNormal   = polygon[polNo].normal[0] =
	polygon[polNo].normal[1] = polygon[polNo].normal[2] =
		UnitNormal(polygon[polNo].vertex[1] - polygon[polNo].vertex[0],
			   polygon[polNo].vertex[2] - polygon[polNo].vertex[1]) ;

	return *this ;
}

Object& Object::MakePolygon(int polNo, int v1, int v2, int v3, int v4, const GLfloat data[][3], const GLfloat m[16])
{
	polygon[polNo].vertex[0].X() = data[v1][0] ;
	polygon[polNo].vertex[0].Y() = data[v1][1] ;
	polygon[polNo].vertex[0].Z() = data[v1][2] ;
	if (m) polygon[polNo].vertex[0].Transform(m) ;

	polygon[polNo].vertex[1].X() = data[v2][0] ;
	polygon[polNo].vertex[1].Y() = data[v2][1] ;
	polygon[polNo].vertex[1].Z() = data[v2][2] ;
	if (m) polygon[polNo].vertex[1].Transform(m) ;

	polygon[polNo].vertex[2].X() = data[v3][0] ;
	polygon[polNo].vertex[2].Y() = data[v3][1] ;
	polygon[polNo].vertex[2].Z() = data[v3][2] ;
	if (m) polygon[polNo].vertex[2].Transform(m) ;

	polygon[polNo].vertex[3].X() = data[v4][0] ;
	polygon[polNo].vertex[3].Y() = data[v4][1] ;
	polygon[polNo].vertex[3].Z() = data[v4][2] ;
	if (m) polygon[polNo].vertex[3].Transform(m) ;

	polygon[polNo].pNormal   = polygon[polNo].normal[0] =
	polygon[polNo].normal[1] = polygon[polNo].normal[2] =
	polygon[polNo].normal[3] = 
		UnitNormal(polygon[polNo].vertex[1] - polygon[polNo].vertex[0],
			   polygon[polNo].vertex[2] - polygon[polNo].vertex[1]) ;

	return *this ;
}

Object& Object::AverageNormal(int pNo, int vNo, int pNo1, int pNo2, int pNo3, int pNo4, int pNo5, int pNo6)
{
	Vector3d sum = polygon[pNo1].pNormal + polygon[pNo2].pNormal ;

	if (pNo3 >= 0) sum += polygon[pNo3].pNormal ;
	if (pNo4 >= 0) sum += polygon[pNo4].pNormal ;
	if (pNo5 >= 0) sum += polygon[pNo5].pNormal ;
	if (pNo6 >= 0) sum += polygon[pNo6].pNormal ;

	polygon[pNo].normal[vNo] = Unit(sum) ;

	return *this ;
}

Object& Object::DrawTriangle(int p, int depth, int average, int flatFlag)
{
	int      i, pFlag = TRUE ;	// 法線共有フラグ
	GLfloat vtx[6][3] ;
	Object   obj(5) ;

	// フラット指定または、各頂点の法線が全て同じ場合
	if (flatFlag || (polygon[p].normal[0] == polygon[p].normal[1] &&
					 polygon[p].normal[1] == polygon[p].normal[2])) pFlag = FALSE ;

	if (depth <= 0)
	{
		if (flatFlag)	// フラット時は、DrawTriangles() で、glBegin() 済み
		{
			glNormal3fv(polygon[p].normal[0].v) ;

			glVertex3fv(polygon[p].vertex[0].v) ;
			glVertex3fv(polygon[p].vertex[1].v) ;
			glVertex3fv(polygon[p].vertex[2].v) ;
		}
		else	// フラット指定なし時（glBegin(), glNormal() 必要）
		{
			// ここまで来るのは、法線が全て同じポリゴンか
			// 初めから depth = 0 に指定されたもの
			// 法線が違う場合（スムーズシェーディング用）は、基本的に depth == 1 の時に処理）
			glBegin(GL_TRIANGLES) ;
				glNormal3fv(polygon[p].normal[0].v) ;
				glVertex3fv(polygon[p].vertex[0].v) ;

				// 最初から depth == 0 のものは、法線が頂点毎に違う可能性あり
				if (pFlag) glNormal3fv(polygon[p].normal[1].v) ;
				glVertex3fv(polygon[p].vertex[1].v) ;

				if (pFlag) glNormal3fv(polygon[p].normal[2].v) ;
				glVertex3fv(polygon[p].vertex[2].v) ;
			glEnd() ;
		}
		return *this ;
	}

	// ポリゴン細分化処理

	VectorToArray(vtx[0], polygon[p].vertex[0]) ;
	VectorToArray(vtx[1], polygon[p].vertex[1]) ;
	VectorToArray(vtx[2], polygon[p].vertex[2]) ;

	if (average)
	{
		float a3 = (polygon[p].vertex[0].Abs() + polygon[p].vertex[1].Abs()) / 2.0f,
			   a4 = (polygon[p].vertex[1].Abs() + polygon[p].vertex[2].Abs()) / 2.0f,
			   a5 = (polygon[p].vertex[2].Abs() + polygon[p].vertex[0].Abs()) / 2.0f ;

		VectorToArray(vtx[3], Unit(polygon[p].vertex[0] + polygon[p].vertex[1]) * a3) ;
		VectorToArray(vtx[4], Unit(polygon[p].vertex[1] + polygon[p].vertex[2]) * a4) ;
		VectorToArray(vtx[5], Unit(polygon[p].vertex[2] + polygon[p].vertex[0]) * a5) ;
	}
	else
	{

		VectorToArray(vtx[3], (polygon[p].vertex[0] + polygon[p].vertex[1]) / 2.0f) ;
		VectorToArray(vtx[4], (polygon[p].vertex[1] + polygon[p].vertex[2]) / 2.0f) ;
		VectorToArray(vtx[5], (polygon[p].vertex[2] + polygon[p].vertex[0]) / 2.0f) ;
	}

	obj.MakePolygon(0, 0,3,5, vtx) ;
	obj.MakePolygon(1, 3,1,4, vtx) ;
	obj.MakePolygon(2, 5,4,2, vtx) ;
	obj.MakePolygon(3, 3,4,5, vtx) ;

	// 分割前のポリゴンの法線が頂点毎に違う場合は平均化した法線を計算
	// 同じだった場合はポリゴン生成時に計算された法線をそのまま使用
	if (pFlag)
	{
		obj.polygon[0].normal[0] = polygon[p].normal[0] ;
		obj.polygon[1].normal[1] = polygon[p].normal[1] ;
		obj.polygon[2].normal[2] = polygon[p].normal[2] ;

		obj.polygon[0].normal[1] = obj.polygon[1].normal[0] = obj.polygon[3].normal[0] =
			Unit(polygon[p].normal[0] + polygon[p].normal[1]) ;
		obj.polygon[1].normal[2] = obj.polygon[2].normal[1] = obj.polygon[3].normal[1] =
			Unit(polygon[p].normal[1] + polygon[p].normal[2]) ;
		obj.polygon[0].normal[2] = obj.polygon[2].normal[0] = obj.polygon[3].normal[2] =
			Unit(polygon[p].normal[2] + polygon[p].normal[0]) ;
	}

	// depth == 1 で、flatFlag 指定がなく
	// なおかつ分割後の法線が隣のポリゴンと共有している場合（!average || pFlag）
	if (depth == 1 && !flatFlag && (!average || pFlag))
	{
// EXT_vertex_array がサポートされておらず、
// OpenGL-1.0 または DONOT_USE_GL_1_1_VERTEX_ARRAY が定義されている場合
//#if !defined GL_EXT_vertex_array && (!defined GL_VERSION_1_1 || defined DONOT_USE_GL_1_1_VERTEX_ARRAY)

		glBegin(GL_TRIANGLES) ;

			for (int p = 0 ; p < 4 ; p ++)
			{
				glNormal3fv(obj.polygon[p].normal[0].v) ;
				glVertex3fv(obj.polygon[p].vertex[0].v) ;

				if (pFlag) glNormal3fv(obj.polygon[p].normal[1].v) ;
				glVertex3fv(obj.polygon[p].vertex[1].v) ;

				if (pFlag) glNormal3fv(obj.polygon[p].normal[2].v) ;
				glVertex3fv(obj.polygon[p].vertex[2].v) ;
			      }
		glEnd() ;
/*
		// GL_TRIANGLE_STRIP で３枚（0, 3, 1）+ GL_TRIANGLES で１枚（2）

		glBegin(GL_TRIANGLE_STRIP) ;

			glNormal3fv(obj.polygon[1].normal[1].v) ;
			glVertex3fv(polygon[1].vertex[1].v) ;

			// 最初の頂点以外の法線は、法線の平均化を行っていない場合は必要なし
			if (pFlag) glNormal3fv(obj.polygon[1].normal[2].v) ;
			glVertex3fv(obj.polygon[1].vertex[2].v) ;

			if (pFlag) glNormal3fv(obj.polygon[0].normal[1].v) ;
			glVertex3fv(obj.polygon[0].vertex[1].v) ;

			if (pFlag) glNormal3fv(obj.polygon[0].normal[2].v) ;
			glVertex3fv(obj.polygon[0].vertex[2].v) ;

			if (pFlag) glNormal3fv(obj.polygon[0].normal[0].v) ;
			glVertex3fv(obj.polygon[0].vertex[0].v) ;

		glEnd() ;

		glBegin(GL_TRIANGLES) ;

			glNormal3fv(obj.polygon[0].normal[0].v) ;
			glVertex3fv(obj.polygon[0].vertex[0].v) ;

			if (pFlag) glNormal3fv(obj.polygon[0].normal[1].v) ;
			glVertex3fv(obj.polygon[0].vertex[1].v) ;

			if (pFlag) glNormal3fv(obj.polygon[0].normal[2].v) ;
			glVertex3fv(obj.polygon[0].vertex[2].v) ;

		glEnd() ;
*/
/*
#else // OpenGL-1.1、または GL_EXT_vertex_array が使用可能な場合

		Vector3d normals[6] ;

		normals[0] = obj.polygon[0].normal[0] ;
		normals[1] = obj.polygon[1].normal[1] ;
		normals[2] = obj.polygon[2].normal[2] ;
		normals[3] = obj.polygon[0].normal[1] ;
		normals[4] = obj.polygon[1].normal[2] ;
		normals[5] = obj.polygon[0].normal[2] ;

// OpenGL-1.0 で GL_EXT_vertex_array がサポートされているか、
// DONOT_USE_GL_1_1_VERTEX_ARRAY が定義されている場合
#if (!defined GL_VERSION_1_1 && defined GL_EXT_vertex_array) || defined DONOT_USE_GL_1_1_VERTEX_ARRAY

		glEnable(GL_VERTEX_ARRAY_EXT) ;
		glEnable(GL_NORMAL_ARRAY_EXT) ;

DONOT_USE_GL_1_1_VERTEX_ARRAY;

		glVertexPointerEXT(3, GL_FLOAT, 0, 6, (GLfloat *)(&vtx[0][0])) ;
		glNormalPointerEXT(GL_FLOAT, sizeof(Vector3d), 6, (GLfloat *)(&normals[0].v[0])) ;

		glBegin(GL_TRIANGLES) ;
			glArrayElement(0) ;
			glArrayElement(3) ;
			glArrayElement(5) ;

			glArrayElement(3) ;
			glArrayElement(1) ;
			glArrayElement(4) ;

			glArrayElement(5) ;
			glArrayElement(4) ;
			glArrayElement(2) ;

			glArrayElement(3) ;
			glArrayElement(4) ;
			glArrayElement(5) ;
		glEnd() ;

#else

		GLubyte indices[] = { 0, 3, 5,  3, 1, 4,  5, 4, 2,  3, 4, 5 } ;

		glEnableClientState(GL_VERTEX_ARRAY) ;
		glEnableClientState(GL_NORMAL_ARRAY) ;

		glVertexPointer(3, GL_FLOAT, 0, (GLfloat *)(&vtx[0][0])) ;
		glNormalPointer(GL_FLOAT, sizeof(Vector3d), (GLfloat *)(&normals[0].v[0])) ;

		glDrawElements(GL_TRIANGLES, 12, GL_UNSIGNED_BYTE, indices) ;

#endif
#endif
*/
		return *this ;
	}

	for (i = 0 ; i < 4 ; i ++) obj.DrawTriangle(i, depth - 1, average, flatFlag) ;

	return *this ;
}

// flatFlag … フラットシェーディングと仮定
// 各ポリゴンを構成する頂点が全ておなじ法線を持つもの仮定して描画
Object& Object::DrawTriangles(int depth, int average, int flatFlag)
{
	int i ;

	// 細分化なしの場合は、平均化もなし
	if (depth <= 0) average = FALSE ;

	// フラットシェーディングと仮定された場合（つまり各ポリゴンの頂点毎の法線がすべて同じ）で
	if (flatFlag)
	{
		glBegin(GL_TRIANGLES) ;

			for (i = 0 ; i < nPolygons ; i ++)
			{
				// フラットシェーディングで頂点座標を平均化しない場合は、最初に法線の０番を指定
				if (!average) glNormal3fv(polygon[i].normal[0].v) ;

				// 三角ポリゴン描画
				DrawTriangle(i, depth, average, TRUE /* flatFlag */) ;
			}

		glEnd() ;
	}
	else	// 通常の分割処理の場合
		for (i = 0 ; i < nPolygons ; i ++) DrawTriangle(i, depth, average) ;

	return *this ;
}


RollingObject& RollingObject::MakeRollingPolygons(int u, int v, GLfloat data[][2], GLfloat x, GLfloat y, GLfloat z)
{
	int      i ;
	GLfloat (*vertex)[3] ;
	if (u < 3 || v < 2) return *this ;

	vertex = new GLfloat[v * 2][3] ;

	uDiv = u ;
	vDiv = v ;

	if (data[    0][0] == 0.0f) fType = 1 ;
	else                       fType = 0 ;
	if (data[v - 1][0] == 0.0f) eType = 1 ;
	else                       eType = 0 ;

	for (i = 0 ; i < u ; i ++)
	{
		int    j, pNo = i * (v - 1) ;
		float rad = 2.0f * PAI / u * i, pRad = 2.0f * PAI / u * (i - 1) ;

		for (j = 0 ; j < v ; j ++)
		{
			vertex[j * 2    ][0] = data[j][0] * x * sin( rad) ;
			vertex[j * 2    ][1] = data[j][0] * y * cos( rad) ;
			vertex[j * 2    ][2] = data[j][1] * z ;

			vertex[j * 2 + 1][0] = data[j][0] * x * sin(pRad) ;
			vertex[j * 2 + 1][1] = data[j][0] * y * cos(pRad) ;
			vertex[j * 2 + 1][2] = data[j][1] * z ;
		}

		if (fType) MakePolygon(pNo        , 0          , 2          , 3        , vertex) ;
		if (eType) MakePolygon(pNo + v - 2, (v - 2) * 2, (v - 1) * 2, v * 2 - 3, vertex) ;

		for (j = fType ; j < v - 1 - eType ; j ++)
			MakePolygon(pNo + j, j * 2, j * 2 + 2, j * 2 + 3, j * 2 + 1, vertex) ;
	}

	delete [] vertex ;
	return *this ;
}

int RollingObject::VFlag(int v, int ud, int vFlag, int vFlags[][2])
{
	if (vFlag)
	{
		if (vFlags) return vFlags[v][ud] ;
		else        return TRUE ;
	}
	else            return FALSE ;
}

RollingObject& RollingObject::AverageNormals(int uFlag, int vFlag, int vFlags[][2]) //, int vNo)
{
	int u ;
	if (uDiv < 3 || vDiv < 2) return *this ;

	for (u = 0 ; u < uDiv ; u ++)
	{
		int v, pNo = u * (vDiv - 1), pPNo = (u - 1) * (vDiv - 1), nPNo = (u + 1) * (vDiv - 1) ;

		if      (u ==        0) pPNo = (uDiv - 1) * (vDiv - 1) ;
		else if (u == uDiv - 1) nPNo = 0 ;

		for (v = 0 /*vNo*/ ; v < vDiv - 1 ; v ++)
		{
			int p = pNo + v, pp = pPNo + v, np = nPNo + v ;

			// Vertex 0
			if (!v && uFlag)
			{
				if (fType) AverageNormal(p, 0, p,np, pp) ;
				else       AverageNormal(p, 0, p,np) ;
			}
			else if (uFlag)
			{
				if (VFlag(v, 0, vFlag, vFlags)) AverageNormal(p, 0, p,np,p-1,np-1) ;
				else                            AverageNormal(p, 0, p,np) ;
			}
			else if (!uFlag && VFlag(v, 0, vFlag, vFlags)) AverageNormal(p, 0, p,p-1) ;

			// Vertex 1
			if (v == vDiv - 2 && uFlag)
			{
				if (eType) AverageNormal(p, 1, p,np,pp) ;
				else       AverageNormal(p, 1, p,np) ;
			}
			else if (uFlag)
			{
				if (VFlag(v, 1, vFlag, vFlags)) AverageNormal(p, 1, p,np,p+1,np+1) ;
				else                            AverageNormal(p, 1, p,np) ;
			}
			else if (!uFlag && VFlag(v, 1, vFlag, vFlags)) AverageNormal(p, 1, p,p+1) ;

			// Vertex 2
			if (v == vDiv - 2 && uFlag)
			{
				if (eType && VFlag(v, 1, vFlag, vFlags)) AverageNormal(p, 2, p,pp,p-1,pp-1) ;
				else                                     AverageNormal(p, 2, p,pp) ;
			}
			else if (uFlag)
			{
				if (VFlag(v, 1, vFlag, vFlags)) AverageNormal(p, 2, p,pp,p+1,pp+1) ;
				else                            AverageNormal(p, 2, p,pp) ;
			}
			else if (!uFlag && VFlag(v, 1, vFlag, vFlags)) AverageNormal(p, 2, p,p+1) ;

			// Vertex 3
			if (!v && uFlag) { if (!fType) AverageNormal(p, 3, p,pp) ; }
			else if (uFlag)
			{
				if (VFlag(v, 0, vFlag, vFlags)) AverageNormal(p, 3, p,pp,p-1,pp-1) ;
				else                            AverageNormal(p, 3, p,pp) ;
			}
			else if (!uFlag && VFlag(v, 0, vFlag, vFlags)) AverageNormal(p, 3, p,p-1) ;
		}
	}
	return *this ;
}

RollingObject& RollingObject::TopNormal(GLfloat x, GLfloat y, GLfloat z)
{
	if (!fType) return *this ;

	int u ;
	for (u = 0 ; u < uDiv ; u ++)
	{
		int pNo = u * (vDiv - 1) ;

		polygon[pNo].normal[0].v[0] = x ;
		polygon[pNo].normal[0].v[1] = y ;
		polygon[pNo].normal[0].v[2] = z ;
	}

	return *this ;
}

RollingObject& RollingObject::BottomNormal(GLfloat x, GLfloat y, GLfloat z)
{
	if (!eType) return *this ;

	int u ;
	for (u = 0 ; u < uDiv ; u ++)
	{
		int pNo = u * (vDiv - 1) + vDiv - 2 ;

		polygon[pNo].normal[1].v[0] = x ;
		polygon[pNo].normal[1].v[1] = y ;
		polygon[pNo].normal[1].v[2] = z ;
	}

	return *this ;
}


// 回転体オブジェクトの描画（法線、頂点座標）
void RollingObject::DrawObject()
{
	int u, v ;

	for (v = 0 ; v < vDiv - 1 ; v ++)
	{
		if (v == 0 && fType) // てっぺんの三角形
		{
			int pNo = (uDiv - 1) * (vDiv - 1) ;

			glBegin(GL_TRIANGLE_FAN) ;

				glNormal3fv(polygon[pNo].normal[0].v) ;
				glVertex3fv(polygon[pNo].vertex[0].v) ;

				glNormal3fv(polygon[pNo].normal[1].v) ;
				glVertex3fv(polygon[pNo].vertex[1].v) ;

				for (u = uDiv - 1 ; u >= 0 ; u --)
				{
					glNormal3fv(polygon[pNo].normal[2].v) ;
					glVertex3fv(polygon[pNo].vertex[2].v) ;

					pNo -= vDiv - 1 ;
				}

			glEnd() ;
		}
		else if (v == vDiv - 2 && eType) // 底の三角形
		{
			int pNo = v ;

			glBegin(GL_TRIANGLE_FAN) ;

				glNormal3fv(polygon[pNo].normal[1].v) ;
				glVertex3fv(polygon[pNo].vertex[1].v) ;

				glNormal3fv(polygon[pNo].normal[2].v) ;
				glVertex3fv(polygon[pNo].vertex[2].v) ;

				for (u = 0 ; u < uDiv ; u ++)
				{
					glNormal3fv(polygon[pNo].normal[0].v) ;
					glVertex3fv(polygon[pNo].vertex[0].v) ;

					pNo += vDiv - 1 ;
				}

			glEnd() ;
		}
		else  // 通常の四角形
		{
			int pNo = v ;

			glBegin(GL_QUAD_STRIP) ;

				glNormal3fv(polygon[pNo].normal[2].v) ;
				glVertex3fv(polygon[pNo].vertex[2].v) ;

				glNormal3fv(polygon[pNo].normal[3].v) ;
				glVertex3fv(polygon[pNo].vertex[3].v) ;

				for (u = 0 ; u < uDiv ; u ++)
				{
					glNormal3fv(polygon[pNo].normal[1].v) ;
					glVertex3fv(polygon[pNo].vertex[1].v) ;

					glNormal3fv(polygon[pNo].normal[0].v) ;
					glVertex3fv(polygon[pNo].vertex[0].v) ;

					pNo += vDiv - 1 ;
				}

			glEnd() ;
		}
	}
}


PositivePolyhedron& PositivePolyhedron::Smooth()
{
	int i, j ;
	
	for (i = 0 ; i < nPolygons ; i ++)
		for (j = 0 ; j < 3 ; j ++)
			polygon[i].normal[j] = Unit(polygon[i].vertex[j]) ;

	return *this ;
}


Octahedron& Octahedron::Make(GLfloat x, GLfloat y, GLfloat z, int d, int a, int smooth)
{
	int      i ;
	GLfloat m[] = {   x, 0.0f, 0.0f, 0.0f,
					 0.0f,   y, 0.0f, 0.0f,
					 0.0f, 0.0f,   z, 0.0f,
					 0.0f, 0.0f, 0.0f, 1.0f } ;

	if (d >= 0) depth   = d ;
	if (a >= 0) average = a ;

	NPolygons(8) ;
	for (i = 0 ; i < 8 ; i ++) MakePolygon(i, o_o[i][0],o_o[i][1],o_o[i][2], o_v, m) ;

	if (smooth) Smooth() ;

	return *this ;
}

void Octahedron::DrawPolyhedron(GLfloat r)
{
	int     i ;
	GLfloat v[6][3] ;

	for (i = 0 ; i < 6 ; i ++)
	{
		v[i][0] = o_v[i][0] * r ;
		v[i][1] = o_v[i][1] * r ;
		v[i][2] = o_v[i][2] * r ;
	}

	Octahedron obj(r, r, r) ;
/*
#if defined GL_VERSION_1_1 && !defined DONOT_USE_GL_1_1_VERTEX_ARRAY
	glEnableClientState(GL_VERTEX_ARRAY) ;
	glVertexPointer(3, GL_FLOAT, 0, (GLfloat *)(&v[0][0])) ;
#endif
*/
	glBegin(GL_TRIANGLES) ;
		for (i = 0 ; i < 8 ; i ++)
		{
			glNormal3fv(obj.polygon[i].normal[0].v) ;
/*
#if defined GL_VERSION_1_1 && !defined DONOT_USE_GL_1_1_VERTEX_ARRAY
			glArrayElement(o_o[i][0]) ;
			glArrayElement(o_o[i][1]) ;
			glArrayElement(o_o[i][2]) ;
#else
*/
			glVertex3fv(v[o_o[i][0]]) ;
			glVertex3fv(v[o_o[i][1]]) ;
			glVertex3fv(v[o_o[i][2]]) ;
/*
#endif
*/
		}
	glEnd() ;
}

GLfloat Octahedron::o_v[][3] = { {  0.0f, 1.0f, 0.0f }, { -1.0f, 0.0f, 0.0f }, {  0.0f,-1.0f, 0.0f },
								 {  1.0f, 0.0f, 0.0f }, {  0.0f, 0.0f, 1.0f }, {  0.0f, 0.0f,-1.0f } } ;
int Octahedron::o_o[][3] = { { 0, 1, 4 }, { 1, 2, 4 }, { 2, 3, 4 }, { 3, 0, 4 },
							   { 1, 0, 5 }, { 2, 1, 5 }, { 3, 2, 5 }, { 0, 3, 5 } } ;


Icosahedron& Icosahedron::Make(GLfloat x, GLfloat y, GLfloat z, int d, int a, int smooth)
{
	int      i ;
	GLfloat m[] = {   x, 0.0f, 0.0f, 0.0f,
					 0.0f,   y, 0.0f, 0.0f,
					 0.0f, 0.0f,   z, 0.0f,
					 0.0f, 0.0f, 0.0f, 1.0f } ;

	if (d >= 0) depth   = d ;
	if (a >= 0) average = a ;

	NPolygons(20) ;
	for (i = 0 ; i < 20 ; i ++) MakePolygon(i, i_o[i][2],i_o[i][1],i_o[i][0], i_v, m) ;

	if (smooth) Smooth() ;

	return *this ;
}

void Icosahedron::DrawPolyhedron(GLfloat r)
{
	int     i ;
	GLfloat v[12][3] ;

	for (i = 0 ; i < 12 ; i ++)
	{
		v[i][0] = i_v[i][0] * r ;
		v[i][1] = i_v[i][1] * r ;
		v[i][2] = i_v[i][2] * r ;
	}

	Icosahedron obj(r, r, r) ;
/*
#if defined GL_VERSION_1_1 && !defined DONOT_USE_GL_1_1_VERTEX_ARRAY
	glEnableClientState(GL_VERTEX_ARRAY) ;
	glVertexPointer(3, GL_FLOAT, 0, (GLfloat *)(&v[0][0])) ;
#endif
*/
	glBegin(GL_TRIANGLES) ;
		for (i = 0 ; i < 20 ; i ++)
		{
			glNormal3fv(obj.polygon[i].normal[0].v) ;
/*
#if defined GL_VERSION_1_1 && !defined DONOT_USE_GL_1_1_VERTEX_ARRAY
			glArrayElement(i_o[i][2]) ;
			glArrayElement(i_o[i][1]) ;
			glArrayElement(i_o[i][0]) ;
#else
*/
			glVertex3fv(v[i_o[i][2]]) ;
			glVertex3fv(v[i_o[i][1]]) ;
			glVertex3fv(v[i_o[i][0]]) ;
/*
#endif
*/
		}
	glEnd() ;
}


// 屈折描画（屈折関係データ計算、OpenGL 描画処理）
// 環境マップ描画はまた別
// glEnable(GL_TEXTURE_2D) ;
// glDisable(GL_LIGHTING) ;
RefractRollingObject& RefractRollingObject::DrawRefractObject(const Vector3d& r_c, const Vector3d& s_c)
{
	int     u, v ;
	GLfloat matrix[16] ;

	// 現在のモデルビュー行列ゲット
	glGetFloatv(GL_MODELVIEW_MATRIX, matrix) ;

	// 頂点単位の屈折データ（テクスチャ座標、透過、反射色計算）
	Vertex(0, 0).CalculateVertex(matrix, r_c, s_c) ; // てっぺんの頂点
	Vertex(0, vDiv - 1).CalculateVertex(matrix, r_c, s_c) ; // 底の頂点

	// それ以外（通常の回転頂点）
	for (v = 0 ; v < vDiv - 1 ; v ++)
	{
		for (u = 0 ; u < uDiv ; u ++)
			Vertex(u, v).CalculateVertex(matrix, r_c, s_c) ;

		// 一周して戻ってきた頂点は最初と同じ
		Vertex(uDiv, v) = Vertex(0, v) ;
	}

	// OpenGL 描画

	// てっぺんの三角形ファン
	glBegin(GL_TRIANGLE_FAN) ;
	{
		glColor3fv(Vertex(0, 0).refraction_color.v) ;
		glTexCoord2fv(Vertex(0, 0).texture.v) ;
		glVertex3fv(Vertex(0, 0).vertex.v) ;

		for (u = 0 ; u <= uDiv ; u ++)
		{
			glColor3fv(Vertex(u, 1).refraction_color.v) ;
			glTexCoord2fv(Vertex(u, 1).texture.v) ;
			glVertex3fv(Vertex(u, 1).vertex.v) ;
		}
	}
	glEnd() ;

	// 底の三角形ファン
	glBegin(GL_TRIANGLE_FAN) ;
	{
		glColor3fv(Vertex(0, vDiv - 1).refraction_color.v) ;
		glTexCoord2fv(Vertex(0, vDiv - 1).texture.v) ;
		glVertex3fv(Vertex(0, vDiv - 1).vertex.v) ;

		for (u = uDiv ; u >= 0 ; u --)
		{
			glColor3fv(Vertex(u, vDiv - 2).refraction_color.v) ;
			glTexCoord2fv(Vertex(u, vDiv - 2).texture.v) ;
			glVertex3fv(Vertex(u, vDiv - 2).vertex.v) ;
		}
	}
	glEnd() ;

	// それ以外（通常の QUAD_STRIP）
	for (v = 1 ; v < vDiv - 1 ; v ++)
	{
		glBegin(GL_QUAD_STRIP) ;
		{
			for (u = 0 ; u <= uDiv ; u ++)
			{
				// 上
				glColor3fv(Vertex(u, v).refraction_color.v) ;
				glTexCoord2fv(Vertex(u, v).texture.v) ;
				glVertex3fv(Vertex(u, v).vertex.v) ;
				// 下
				glColor3fv(Vertex(u, v + 1).refraction_color.v) ;
				glTexCoord2fv(Vertex(u, v + 1).texture.v) ;
				glVertex3fv(Vertex(u, v + 1).vertex.v) ;
			}
		}
		glEnd() ;
	}

	return *this ;
}

// 屈折描画（屈折関係データ計算、OpenGL 描画処理）
// 環境マップ描画はまた別
// 透過率、反射率は計算しない（アプリケーションが自前で glColor() を実行）
// glEnable(GL_TEXTURE_2D) ;
// glDisable(GL_LIGHTING) ;
// glColor*() ;
RefractRollingObject& RefractRollingObject::DrawRefractObject()
{
	int     u, v ;
	GLfloat matrix[16] ;

	// 現在のモデルビュー行列ゲット
	glGetFloatv(GL_MODELVIEW_MATRIX, matrix) ;

	// 頂点単位の屈折データ（テクスチャ座標、透過、反射色計算）
	Vertex(0, 0).CalculateVertex(matrix) ; // てっぺんの頂点
	Vertex(0, vDiv - 1).CalculateVertex(matrix) ; // 底の頂点

	// それ以外（通常の回転頂点）
	for (v = 0 ; v < vDiv - 1 ; v ++)
	{
		for (u = 0 ; u < uDiv ; u ++)
			Vertex(u, v).CalculateVertex(matrix) ;

		// 一周して戻ってきた頂点は最初と同じ
		Vertex(uDiv, v) = Vertex(0, v) ;
	}

	// OpenGL 描画

	// てっぺんの三角形ファン
	glBegin(GL_TRIANGLE_FAN) ;
	{
		glTexCoord2fv(Vertex(0, 0).texture.v) ;
		glVertex3fv(Vertex(0, 0).vertex.v) ;

		for (u = 0 ; u <= uDiv ; u ++)
		{
			glTexCoord2fv(Vertex(u, 1).texture.v) ;
			glVertex3fv(Vertex(u, 1).vertex.v) ;
		}
	}
	glEnd() ;

	// 底の三角形ファン
	glBegin(GL_TRIANGLE_FAN) ;
	{
		glTexCoord2fv(Vertex(0, vDiv - 1).texture.v) ;
		glVertex3fv(Vertex(0, vDiv - 1).vertex.v) ;

		for (u = uDiv ; u >= 0 ; u --)
		{
			glTexCoord2fv(Vertex(u, vDiv - 2).texture.v) ;
			glVertex3fv(Vertex(u, vDiv - 2).vertex.v) ;
		}
	}
	glEnd() ;

	// それ以外（通常の QUAD_STRIP）
	for (v = 1 ; v < vDiv - 2 ; v ++)
	{
		glBegin(GL_QUAD_STRIP) ;
		{
			for (u = 0 ; u <= uDiv ; u ++)
			{
				// 上
				glTexCoord2fv(Vertex(u, v).texture.v) ;
				glVertex3fv(Vertex(u, v).vertex.v) ;
				// 下
				glTexCoord2fv(Vertex(u, v + 1).texture.v) ;
				glVertex3fv(Vertex(u, v + 1).vertex.v) ;
			}
		}
		glEnd() ;
	}

	return *this ;
}


// 屈折描画（我流フェイク屈折計算、OpenGL 描画処理）
// 環境マップ描画はまた別
// glEnable(GL_TEXTURE_2D) ;
// glDisable(GL_LIGHTING) ;
RefractRollingObject& RefractRollingObject::DrawFakeRefractObject(const Vector3d& r_c, const Vector3d& s_c)
{
	int     u, v ;
	GLfloat matrix[16] ;

	// 現在のモデルビュー行列ゲット
	glGetFloatv(GL_MODELVIEW_MATRIX, matrix) ;

	// 頂点単位の屈折データ（テクスチャ座標、透過、反射色計算）
	Vertex(0, 0).CalculateFakeVertex(matrix, r_c, s_c) ; // てっぺんの頂点
	Vertex(0, vDiv - 1).CalculateFakeVertex(matrix, r_c, s_c) ; // 底の頂点

	// それ以外（通常の回転頂点）
	for (v = 0 ; v < vDiv - 1 ; v ++)
	{
		for (u = 0 ; u < uDiv ; u ++)
			Vertex(u, v).CalculateFakeVertex(matrix, r_c, s_c) ;

		// 一周して戻ってきた頂点は最初と同じ
		Vertex(uDiv, v) = Vertex(0, v) ;
	}

	// OpenGL 描画

	// てっぺんの三角形ファン
	glBegin(GL_TRIANGLE_FAN) ;
	{
		glColor3fv(Vertex(0, 0).refraction_color.v) ;
		glTexCoord2fv(Vertex(0, 0).texture.v) ;
		glVertex3fv(Vertex(0, 0).vertex.v) ;

		for (u = 0 ; u <= uDiv ; u ++)
		{
			glColor3fv(Vertex(u, 1).refraction_color.v) ;
			glTexCoord2fv(Vertex(u, 1).texture.v) ;
			glVertex3fv(Vertex(u, 1).vertex.v) ;
		}
	}
	glEnd() ;

	// 底の三角形ファン
	glBegin(GL_TRIANGLE_FAN) ;
	{
		glColor3fv(Vertex(0, vDiv - 1).refraction_color.v) ;
		glTexCoord2fv(Vertex(0, vDiv - 1).texture.v) ;
		glVertex3fv(Vertex(0, vDiv - 1).vertex.v) ;

		for (u = uDiv ; u >= 0 ; u --)
		{
			glColor3fv(Vertex(u, vDiv - 2).refraction_color.v) ;
			glTexCoord2fv(Vertex(u, vDiv - 2).texture.v) ;
			glVertex3fv(Vertex(u, vDiv - 2).vertex.v) ;
		}
	}
	glEnd() ;

	// それ以外（通常の QUAD_STRIP）
	for (v = 1 ; v < vDiv - 1 ; v ++)
	{
		glBegin(GL_QUAD_STRIP) ;
		{
			for (u = 0 ; u <= uDiv ; u ++)
			{
				// 上
				glColor3fv(Vertex(u, v).refraction_color.v) ;
				glTexCoord2fv(Vertex(u, v).texture.v) ;
				glVertex3fv(Vertex(u, v).vertex.v) ;
				// 下
				glColor3fv(Vertex(u, v + 1).refraction_color.v) ;
				glTexCoord2fv(Vertex(u, v + 1).texture.v) ;
				glVertex3fv(Vertex(u, v + 1).vertex.v) ;
			}
		}
		glEnd() ;
	}

	return *this ;
}


// 屈折描画（屈折関係データ計算、OpenGL 描画処理）
// 環境マップ描画はまた別
// 透過率、反射率は計算しない（アプリケーションが自前で glColor() を実行）
// glEnable(GL_TEXTURE_2D) ;
// glDisable(GL_LIGHTING) ;
// glColor*() ;
RefractRollingObject& RefractRollingObject::DrawFakeRefractObject()
{
	int     u, v ;
	GLfloat matrix[16] ;

	// 現在のモデルビュー行列ゲット
	glGetFloatv(GL_MODELVIEW_MATRIX, matrix) ;

	// 頂点単位の屈折データ（テクスチャ座標、透過、反射色計算）
	Vertex(0, 0).CalculateFakeVertex(matrix) ; // てっぺんの頂点
	Vertex(0, vDiv - 1).CalculateFakeVertex(matrix) ; // 底の頂点

	// それ以外（通常の回転頂点）
	for (v = 0 ; v < vDiv - 1 ; v ++)
	{
		for (u = 0 ; u < uDiv ; u ++)
			Vertex(u, v).CalculateFakeVertex(matrix) ;

		// 一周して戻ってきた頂点は最初と同じ
		Vertex(uDiv, v) = Vertex(0, v) ;
	}

	// OpenGL 描画

	// てっぺんの三角形ファン
	glBegin(GL_TRIANGLE_FAN) ;
	{
		glTexCoord2fv(Vertex(0, 0).texture.v) ;
		glVertex3fv(Vertex(0, 0).vertex.v) ;

		for (u = 0 ; u <= uDiv ; u ++)
		{
			glTexCoord2fv(Vertex(u, 1).texture.v) ;
			glVertex3fv(Vertex(u, 1).vertex.v) ;
		}
	}
	glEnd() ;

	// 底の三角形ファン
	glBegin(GL_TRIANGLE_FAN) ;
	{
		glTexCoord2fv(Vertex(0, vDiv - 1).texture.v) ;
		glVertex3fv(Vertex(0, vDiv - 1).vertex.v) ;

		for (u = uDiv ; u >= 0 ; u --)
		{
			glTexCoord2fv(Vertex(u, vDiv - 2).texture.v) ;
			glVertex3fv(Vertex(u, vDiv - 2).vertex.v) ;
		}
	}
	glEnd() ;

	// それ以外（通常の QUAD_STRIP）
	for (v = 1 ; v < vDiv - 2 ; v ++)
	{
		glBegin(GL_QUAD_STRIP) ;
		{
			for (u = 0 ; u <= uDiv ; u ++)
			{
				// 上
				glTexCoord2fv(Vertex(u, v).texture.v) ;
				glVertex3fv(Vertex(u, v).vertex.v) ;
				// 下
				glTexCoord2fv(Vertex(u, v + 1).texture.v) ;
				glVertex3fv(Vertex(u, v + 1).vertex.v) ;
			}
		}
		glEnd() ;
	}

	return *this ;
}


// ワイングラス描画（法線、頂点座標）
void WineGlass::DrawObject()
{
	int u, v ;

	for (v = 0 ; v < N_VERTICIES - 1 ; v ++)
	{
		if (v == 0) // てっぺんの三角形
		{
			int pNo = (uDiv - 1) * (N_VERTICIES - 1) ;

			glBegin(GL_TRIANGLE_FAN) ;

				glNormal3fv(polygon[pNo].normal[0].v) ;
//				glNormal3f(0.0, 0.0, 1.0) ;	// 一番上の頂点座標は法線を真上に
				glVertex3fv(polygon[pNo].vertex[0].v) ;

				glNormal3fv(polygon[pNo].normal[1].v) ;
				glVertex3fv(polygon[pNo].vertex[1].v) ;

				for (u = uDiv - 1 ; u >= 0 ; u --)
				{
					glNormal3fv(polygon[pNo].normal[2].v) ;
					glVertex3fv(polygon[pNo].vertex[2].v) ;

					pNo -= N_VERTICIES - 1 ;
				}

			glEnd() ;
		}
		else if (v == N_VERTICIES - 2 ) // 底の三角形
		{
			int pNo = v ;

			glBegin(GL_TRIANGLE_FAN) ;

				glNormal3fv(polygon[pNo].normal[1].v) ;
//				glNormal3d(0.0, 0.0, -1.0) ;	// 一番下の頂点座標は法線を真下に
				glVertex3fv(polygon[pNo].vertex[1].v) ;

				glNormal3fv(polygon[pNo].normal[2].v) ;
				glVertex3fv(polygon[pNo].vertex[2].v) ;

				for (u = 0 ; u < uDiv ; u ++)
				{
					glNormal3fv(polygon[pNo].normal[0].v) ;
					glVertex3fv(polygon[pNo].vertex[0].v) ;

					pNo += N_VERTICIES - 1 ;
				}

			glEnd() ;
		}
		else  // 通常の四角形
		{
			int pNo = v ;

			glBegin(GL_QUAD_STRIP) ;

				glNormal3fv(polygon[pNo].normal[2].v) ;
				glVertex3fv(polygon[pNo].vertex[2].v) ;

				glNormal3fv(polygon[pNo].normal[3].v) ;
				glVertex3fv(polygon[pNo].vertex[3].v) ;

				for (u = 0 ; u < uDiv ; u ++)
				{
					glNormal3fv(polygon[pNo].normal[1].v) ;
					glVertex3fv(polygon[pNo].vertex[1].v) ;

					glNormal3fv(polygon[pNo].normal[0].v) ;
					glVertex3fv(polygon[pNo].vertex[0].v) ;

					pNo += N_VERTICIES - 1 ;
				}

			glEnd() ;
		}
	}
}

void WineGlass::DrawPolyhedron(int divide, GLfloat r)
{
	WineGlass obj(divide, r,r,r) ;
	obj.DrawObject() ;
}


GLfloat Icosahedron::X = 0.525731112119133606f, Icosahedron::Z = 0.850650808352039932f ;
GLfloat Icosahedron::i_v[][3] = { {  -Icosahedron::X, 0.0f,   Icosahedron::Z }, {   Icosahedron::X, 0.0f,   Icosahedron::Z }, {  -Icosahedron::X, 0.0f,  -Icosahedron::Z }, {   Icosahedron::X, 0.0f,  -Icosahedron::Z },
								 { 0.0f,   Icosahedron::Z,   Icosahedron::X }, { 0.0f,   Icosahedron::Z,  -Icosahedron::X }, { 0.0f,  -Icosahedron::Z,   Icosahedron::X }, { 0.0f,  -Icosahedron::Z,  -Icosahedron::X },
								 {   Icosahedron::Z,   Icosahedron::X, 0.0f }, {  -Icosahedron::Z,   Icosahedron::X, 0.0f }, {   Icosahedron::Z,  -Icosahedron::X, 0.0f }, {  -Icosahedron::Z,  -Icosahedron::X, 0.0f } } ;
int Icosahedron::i_o[][3] = { { 0, 4, 1 }, { 0, 9, 4 }, { 9, 5, 4 }, { 4, 5, 8 },{ 4, 8, 1 },
							   { 8,10, 1 }, { 8, 3,10 }, { 5, 3, 8 }, { 5, 2, 3 },{ 2, 7, 3 },
							   { 7,10, 3 }, { 7, 6,10 }, { 7,11, 6 }, {11, 0, 6 },{ 0, 1, 6 },
							   { 6, 1,10 }, { 9, 0,11 }, { 9,11, 2 }, { 9, 2, 5 },{ 7, 2,11 } } ;

GLfloat WineGlass::vertexes[N_VERTICIES][2] =
{
	{ 0.0f, 0.0f }, { 4.0f, 0.2f }, { 8.0f, 1.5f }, { 10.5f, 3.2f }, { 13.0f, 5.8f },
	{ 14.0f, 8.0f }, { 14.2f, 10.0f }, { 14.1f, 12.0f }, { 13.5f, 15.0f }, { 12.2f, 19.0f },
	{ 12.2f, 19.5f }, { 12.3f, 19.7f }, { 12.5f, 19.6f }, { 12.6f, 19.3f }, { 14.0f, 15.0f },
	{ 14.6f, 12.0f }, { 14.8f, 10.0f }, { 14.4f, 7.8f }, { 13.3f, 5.4f }, { 11.2f, 2.8f },
	{ 7.8f, 0.4f }, { 4.0f, -1.1f }, { 3.0f, -1.8f }, { 2.2f, -3.0f }, { 1.8f, -4.2f },
	{ 1.5f, -8.0f }, { 1.5f, -12.0f }, { 1.7f, -15.0f }, { 2.0f, -16.0f }, { 2.7f, -16.7f },
	{ 5.5f, -17.8f }, { 11.0f, -18.7f }, { 12.5f, -19.0f }, { 13.0f, -19.5f }, { 13.0f, -19.8f },
	{ 12.5f, -19.9f }, { 6.0f, -19.9f }, { 3.0f, -19.5f }, { 1.0f, -18.3f }, { 0.0f, -18.0f }
} ;

const Vector3d Vector3d::white(1.0f, 1.0f, 1.0f) ;
