
// よく使う数学関数 & 定数

#ifndef ___MATH3D_H___
#define ___MATH3D_H___

#include <math.h>
#include <stdlib.h>

#include "Define.H"

#ifndef PAI
#undef  PAI
#endif
#define PAI 3.141592653589793238462643383279502884197169399375105820974944

#define INF     1.0e10
#define EPS     1.0e-10

inline int Abs(int n) { return (n >= 0 ? n : -n) ; }
inline double Abs(double d) { return (d >= 0.0 ? d : -d) ; }

inline int Equ(double s, double d) { return (Abs(s - d) < EPS) ; }

inline int Sgn(int n) { return (n > 0 ? 1 : (n < 0 ? -1 : 0)) ; }
inline double Sgn(double d) { return (d>0.0 ? 1.0 : (d<0.0 ? -1.0 : 0.0)) ; }

inline double Square(double d) { return d * d ; }
inline int Square(int d) { return d * d ; }

inline double Cube(double d) { return d * d * d ; }
inline int Cube(int d) { return d * d * d ; }

inline double Pow(double d) { return d * d ; }
inline int Pow(int d) { return d * d ; }

inline double Pow(double d, double n)
{
    return (n == 0.0 ? 1.0 : pow(d, n)) ;
}

inline double Pow(double d, int n)
{
    if (n == 0) return 1.0 ;
    if (d == 0.0) return 0.0 ;

    int     a = Abs(n) ;
    double  m = d ;
    for (int i = 1 ; i < a ; i ++) d *= m ;

    if (n < 0) return 1.0 / d ;
    return d ;
}

inline int Pow(int d, int n)
{
    if (n == 0) return 1 ;
    else if (n < 0) return 0 ;

    int m = d ;
    for (int i = 1 ; i < n ; i ++) d *= m ;
    return d ;
}

// 三角関数関係
inline double Rad(double d) { return (d * PAI / 180.0) ; }
inline double Deg(double r) { return (r * 180.0) / PAI ; }

inline double Sin(double d) { return sin(Rad(d)) ; }
inline double Cos(double d) { return cos(Rad(d)) ; }

inline double ATan2(double y, double x) { return Deg(atan2(y, x)) ; }

inline double Clump(double d, double l = 0.0, double h = 1.0)
{
	if      (d < l) return l ;
	else if (d > h) return h ;
	else            return d ;
}

inline void UnitVector(double v[])
{
	double d ;

	if ((d = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])) != 0.0)
	{
		v[0] /= d ;
		v[1] /= d ;
		v[2] /= d ;
	}
}

inline void NormalVector(double v1[], double v2[], double out[])
{
	out[0] = v1[1]*v2[2] - v1[2]*v2[1] ;
	out[1] = v1[2]*v2[0] - v1[0]*v2[2] ;
	out[2] = v1[0]*v2[1] - v1[1]*v2[0] ;

	UnitVector(out) ;
}


#endif
